--[[
Copyright 2014 by Sean Luke
Licensed under the Apache 2.0 License (found at the end of this file)


HFA.LUA
Sean Luke
Version 1.0
May 2014


A simple package to make it easy to implement hierarchical finite-state
automata or HFA in the form of Moore machines, for purposes of programming
agent behaviors.  This package is not as useful for automata which are intended
for lexing strings.  An HFA is a finite-state automaton whose states correspond
to basic behaviors (a hard-coded behavior you have constructed) or to other 
automata.  While HFA contain other HFA, HFA are not permitted to be recursive: 
that is, you cannot have an HFA which ultimately contains itself.

This package implements all of the features discussed in GMU Computer Science 
Technical Report GMU-CS-TR-2014-5, "Building a Hierarchical Finite-State 
Automaton: A Primer"; and indeed the implementation described in that paper is
a simplified version of this one.  This more extensive version does some more
careful variable existence checking and sets parent pointers to nill.  If you
ever receive a "WARNING" message, this is really a serious error: you've done
something wrong.  Read that primer before continuing with this documentation.


-- Behaviors --

Let's begin with the notion of a BEHAVIOR.  A behavior is a simple object 
(a dictionary) which contains a few things:

1. A NAME (as a string).  This really only exists for debugging (for now).
2. A backpointer to a PARENT of the behavior (or to nil if there is no parent).
   This item may get changed dynamically to various things as the automaton
   is running.  Before any of the functions below (start/stop/go) are called,
   the PARENT will have been set.
3. A START function.  This function is called when the behavior is started.
   This value may be nil.  The function takes two arguments, the first of 
   which is the behavior itself. The second is a table of targets.  We'll 
   get to these later in the TARGETS section, for now set them to NIL.
4. A GO function.  This function is called immediately after START, and
   rapidly / repeatedly thereafter.  This value may be nil.  The function 
   takes a single argument, the first of which is the behavior itself.  
   The second is a table of targets.  We'll get to these later in the TARGETS 
   section, for now set them to NIL.
5. A STOP function.  This function is called when the behavior is stopped.
   This value may be nil.  The function takes a single argument, the first 
   of which is the behavior itself. After being stopped, a behavior can be 
   started again.  The second is a table of targets.  We'll get to these 
   later in the TARGETS section, for now set them to NIL.
5. Whether the behavior has yet been PULSED.  This is used by the pulse function.


Imagine you had a behavior called myBehavior.  Then you could access these as:

myBehavior.name
myBehavior.parent
myBehavior.pulsed
myBehavior.start
myBehavior.stop
myBehavior.go

A BASIC BEHAVIOR is a hard-coded behavior you have created yourself.  You don't
have to implement all three functions (start/stop/go).  For example, if you have
a one-shot behavior (like "kickBall"), you might implement it by simply providing
the start function (to kick the ball) and setting the other functions to nil.
Or if you have a continuous function (like "forward"), you might do this by
providing a go function (to move the robot forward by an epsilon), and set
the others to nil.  Or if you have an asynchronous behavior (like "walk"),
you might implement the start function to tell the robot to begin walking,
set the go function to nil, and set the stop function to tell the robot to stop
walking.  And so on.

A utility function is provided to you to make this easy:

makeBehavior(name, start, stop, go)           returns a behavior.

In most cases you'd store the behavior in a global variable with the same name.  So
if you were creating a behavior called FOO, you'd do so like this:

foo = makeBehavior("foo", ... )
                
                                      
                                 
-- Automata --

An HFA is a behavior associated with a set of SUB-BEHAVIORS in the form of STATES.
For purposes of this library, a state and a behavior are the same thing.  At any 
time, an active HFA has a CURRENT STATE (that is, a current sub-behavior).  There 
is a special initial state called START, which doesn't do anything special at all
(not to be confused with the start function).  Thus we have (for some HFA called 
myHFA):

myHFA.name            (Because an hfa is a behavior)
myHFA.parent          (Likewise)
myHFA.pulsed		  (Likewise)
myHFA.start           (Likewise.  Set to startHFA, see below.)
myHFA.stop            (Likewise.  Set to stopHFA, see below.)
myHFA.go              (Likewise.  Set to goHFA, see below.)
myHFA.current         (The current state)

An HFA is associated with a TRANSITION FUNCTION which takes the HFA as an 
argument and returns what the current state ought to be at this time.  This is
typically done by examining the present setting of the current state and some 
world information to determine the new current state.  More on transition
functions in a bit, but for the moment, note that the transition functino is
stored here:

myHFA.transition

Because an HFA is a behavior, it has a start, stop and go function stored.  All
HFA use exactly the same start, stop, and go functions, and so when you make an HFA
you don't provide them: the HFA set them up themselves.  These special private
functions are called startHFA(...), stopHFA(...), and goHFA(...).

The startHFA(...) function initializes the current state of the HFA to the START
state, among other things.

The stopHFA(...) function sets the current state of the HFA to nil, among other
things.  If the current state wasn't already nil or the "start" state, then its
STOP function is also called recursively.

The goHFA(...) function is where the interesting work is done.  This function first
calls the transition function to determine the new current state.  If the current
state is different from the previous one, then the START function is called on the
new current state, and furthermore if the previous current state wasn't nil and 
wasn't "start", then the previous current state's STOP function is also called.
Regardless, the current state is changed, and GO function is then called on the
new current state.

Notice that in goHFA(...) the transition function is called FIRST, followed by 
the GO function (and possibly START and STOP).  This means that the "start" 
state typically never has its GO function called, ever.  This is intentional.  The "start"
state is just meant to be a dummy state which provides us with a way to define
a transition funtion which, in essence, will specify the initial behavior performed
by the HFA.  It's defined in the global variable:

start = makeBehavior("start", nil, nil, nil)

Also notice that the HFA doesn't actually store any of its sub-behaviors.  They're
just specified by the transition function based on the current state of the HFA.
 
HFA are created with a different utility function than basic hard-coded behaviors.
You call makeHFA, providing the name, the transition function.  

makeHFA(name, transition)           returns an HFA.

Again, when creating an HFA, you'd typically store it in a global variable of the
same name as the HFA, thus if you were creating an HFA called FOO, you'd say:

foo = makeHFA("foo", ... )



-- Transition Functions --

Transition functions for automata can be complex: they subsume all the edges in the
finite-state automaton.  It'd be very inconvenient to have to write a transition
function on your own.  Various items are provided to make this easier on you.

To start, there is utility function makeTransition(...).  This function takes a single
argument, TRANSITIONS, which is a dictionary.  The KEYS of the dictionary
are behaviors.  The VALUES are either BEHAVIORS or are themselves TRANSITION FUNCTIONS
(which, as normal, take the HFA as a single argument and return a desired new current
state).

makeTransition(transitions, nil)                returns a transition function

The function generated by makeTransition(...) works like this.  When called to
determine the next current state, it queries the HFA for the present current state.
It then looks that state up in the dictionary as the key.  If the resulting value is
a behavior, this behavior is returned as the new current state.  If the resulting
value is a transition function, this function is called and its return value is returned
as the new current state.

Thus every state in the HFA can have its own transition function which says where
to go next if you're presently in that state; or it can explicitly declare another state
to unilaterally go to next.  This second case is equivalent to an unlabeled (or "epsilon")
edge in an FSA.

Thus the most common pattern for making an HFA called FOO would be:

foo = makeHFA("foo", makeTransition(...))

For example, you might have something like this.  See if you can make out what's going on:

foo = makeHFA("foo", makeTransition(
    {
    [start] = forward,
    [forward] = function(hfa) if (closeToBall()) then return kick else return forward end end,
    [kick] = rotate,
    [rotate] = function(hfa) if (ballAhead()) then return forward else return rotate end end,
    }))

An additional option is for a transition to be a table of EVENTS (strings) mapping to 
states, plus a state to transition to by default if no events are matched.  This is done
using the utility function events(...) which produces a transition function given
even transitions.  events(...) takes two arguments: (1) a table of states keyed by events and
(2) a default state to transition to.  You might use it like this:

    [kick] = event({ ["move"]=forward, ["success"]=rotate } ),

This means that, when kicking, if the "move" event has been posted, transition to the forward
state.  If the "success" event has been posted, start rotating.  Otherwise, keep on kicking. 
The values passed into event don't have to be just states: they can themselves be transition
functions.

Events can be posted to the hfa by calling postEvent(hfa, event), and are cleared 
at each iteration.  This gizmo makes it easy for you to create 
automata which respond to events essentially by labelling those events on outgoing edges 
to other states.  Though you'd never access this, posted events are stored in

myHFA.events

You can also create states which return events.  In this case, the event is added to
the end of the events list, and may thus trigger a transition.   Since it's at the end of 
the events list, other posted events take precedence.  This event is stored in the following
private variable:

myHFA.goReturned

Here's the order of precedence:

- If the transition is explicitly to another state, then this state will be used.

- Else if there's an explicitly declared transition function for a state, whatever state it
  returns when called will be used (unless it is nil).  This includes a transition
  function built using the event(...) function.  In this case, whatever state
  is associated with it in the table will be used, if any.  Events are considered
  in the order they were posted.
  
- Else no transition will occur.

If you do not specify a transition function, or your transition function does not 
have a transition for a specific current state and that behavior returns nothing, then
no transition is made: the behavior stays the current state.  No warning is issued.



-- Running the HFA --

So you've constructed a hierarchy of automata and basic behaviors.  How do you get the
whole thing going?  With the utility function PULSE(...)

pulse(behavior, targets)    initially calls START on the behavior, then GO.  Thereafter
                            just calls GO on the behavior.
                            
(for now, set targets to nil -- we'll get to them later in the TARGETS section)_

Typically you'd just set up your HFA and then call pulse(...) on it forever.  If for 
some reason you need to reset the HFA after calling pulse(...) on it some number of times,
you can do so with the following function:

reset(behavior, targets)	calls STOP on the behavior, then resets it so pulse will
							call START on it again in the future.

Example:

pulse(myHFA, myTargets)
pulse(myHFA, myTargets)
pulse(myHFA, myTargets)
-- ... and so on, then to reset:
reset(myHFA, myTargets)
-- we're now ready to begin pulsing again...



-- Flags --

HFA come with a variety of gizmos which you will probably find convenient.  Foremost
are FLAGS, which permit a child HFA to inform its parent HFA that it believes it
has completed its task ("done") or that it has "failed".  A flag is set by the
child behavior transitioning to a FLAG BEHAVIOR as its current state.  When this happens,
the corresponding flag is set in the *parent* (not the child).  The parent's transition
functions can test for this flag to determine whether they should now transition away from
the child as the parent's current state, and go somewhere else.  Flags are reset when
startHFA(...) is called on the parent and also when the parent's goHFA(...) function
transitions to a new current state.

There are four flag behaviors available:

done        raises the "done" flag and immediately transitions to the "start" state
failed        raises the "failed" flag and immediately transitions to the "start" state
sayDone        raises the "done" flag
sayFailed    raises the "failed" flag

The first two are the most common, since once you're done (or have failed) you probably
have no further actions you want to do and might as well start over from the top if for
some reason your parent HFA doesn't act on this flag being raised.  For
these two, there's no reason to declare a transition since the transition is
automatic.  For the second two (setDone, setFailed), you need to provide a transition
of some kind, typically a unilateral one.  These second two are much less common.

The flags themselves are stored in the parent here (you'll rarely touch this):

myHFA.done            (false by default)
myHFA.failed        (false by default)

Flags may be PROPAGATED.  It's a common pattern in building HFA that you see transition
functions of the form "if DONE has been rased, then transition to DONE".   This is
essentially saying "If my child thinks he's done, then I'm done too." (likewise "failed").
Because it's so common, rather than require this transition, an HFA can declare that all
of its flags should be propagated to its parent.  This means that if its child sets a flag
in the HFA, the same flag will immediateliy be set in the HFA's parent.  If the parent
likewise has propagation set, it'll be set in the grandparent too, and so on.  If this
feels like throwing an exception, you're on the right track.

You set propagation in your hfa like this:

myHFA.propagateFlags = true

There is an internal function used by the various flag behaviors to set the appropriate
flag and also potentially propagate it recursively:

setFlag(hfa, flag)        sets the given flag in the HFA.  Example: setFlag(myHFA, "failed")



-- COUNTERS AND TIMERS --

Each HFA also has one COUNTER and one TIMER.  HFA nested within one another do not share
these: they have separate counters and timers.

A COUNTER is just an integer which starts at zero (when the HFA's startHFA(...) function is
called) and may be increased or reset to zero by certain behaviors which the HFA may
transition to as its current state.  These behaviors are:

bumpCounter            increments the counter by 1
resetCounter        sets the counter to 0

The current counter value is stored here:

myHFA.counter

... but it's better style to test the current counter value like this:

currentCounter(myHFA)

Testing the value of a counter is useful in a transition function.  Let's say that
you wanted to go forward for three steps and if you were still not close enough
to the ball, you transferred to the run state, then go back to forward.  You might say:

foo = makeHFA("foo", makeTransition(
    {
    [start] = forward,
    [forward] = function(hfa) if (closeToBall()) then return kick else return bumpCounter end end,
    [kick] = rotate,
    [rotate] = function(hfa) if (ballAhead()) then return forward else return rotate end end,
    [bumpCounter] = function(hfa) if (hfa.counter > 3) then return run else return forward end end,
    [run] = resetCounter,
    [resetCounter] = forward,
    }))
    
Similarly, a TIMER is an integer which stores a time interval in seconds.  When the HFA's
startHFA(...) function is called, the timer is set to the current time.  A single behavior
can likewise update the timer to the current time:

resetTimer            updates the timer to the current time

The current timer value is stored here:

myHFA.timer

... but almost certainly you'd access it using the more useful function:

currentTimer(myHFA)

This gives you the current time, minus the timer value, and thus the number of seconds which
have elapsed since the timer was last reset.  Revisiting the example above, imagine you want
to go forward not for three steps but rather for three seconds, then do a run at the ball.  You
might say:

foo = makeHFA("foo", makeTransition(
    {
    [start] = forward,
    [forward] = function(hfa)    
        if (closeToBall()) then return kick 
        elseif  currentTimer(hfa) > 3 then return run
        else return forward 
        end
    end,
    [kick] = rotate,
    [rotate] = function(hfa) if (ballAhead()) then return forward else return rotate end end,
    [run] = resetTimer,
    [resetTimer] = forward,
    }))



-- TARGETS --

It's often useful to parameterize your behaviors, so that you can create a behavior called
GoTo(X) rather than multiple nearly identical behaviors like GoToTheBall and GoToTheGoal
and GoToTheMidfield and so on.  To do this, we introduce the notion of TARGETS.

A Target is a parameter name (a string) bound to a value (a target, whatever you'd like) 
in a table, such as { ["X"] = TheBall }.  When you create a behavior, you can have a table
of this form passed into your START, STOP, and GO functions.  For example, if you were building
a "goto" behavior, you might have its GO function look like this:

-- GOTO(X)
goto = function(hfa, targets)
    position = positionOf(targets.X)
    turnTo(position)
    walkTowards(position)
end

An HFA can of course also have targets.  It is the job of the transitions to map these targets
to the appropriate names used by the sub-behaviors.  For example, imagine if you had an HFA
called ATTACK which expected two targets called "ball" and "goal".  You have a sub-behavior
called goto which needs the goal target, but it doesn't call it the goal, it calls it "X".
When your transition functions return GoTo, they need to also indicate that the "goal" target
should be provided to goto, renamed to "X".  This is done by not returning goto as
the behavior, but rather returning a table of the form:

{ [0] = goto, ["X"] = "goal" }

This says: goto is the behavior, and provide the "goal" target to goto as a target named "X".

In general this table you'd provide looks like:

{ [0] = the behavior,
        ["behavior's name for first target"] = "hfa's name for target", 
        ["behavior's name for second target"] = "hfa's name for second target", 
        ... }
        
Target names should be strings.  Do not use 0 as a target name, that refers to the behavior.
        
For example, we might have:
        
attack = makeHFA("attack", makeTransition(
    {
    [start] = forward,
    [forward] = function(hfa)    
        if (farFromBall()) then return { [0] = goto, ["X"] = "goal" }
        else return kick
        end,
    [goto] = function(hfa)
        if (closeToBall()) then return forward
        else return { [0] = goto, ["X"] = "goal" }
        end,
    [kick] = { [0] = goto, ["X"] = "goal"}, 
    ...
    }))

... Now you have an "attack" behavior which expects a "goal"  paramter (and perhaps
a "ball" paramter used in some other sub-function not shown), and whenever goto is transitioned
to, the "ball" parameter is passed into goal as the "X" parameter.  Note that you can have
sub-behaviors which take all manner of targets.  For example, forward takes no targets.
You could have another sub-behavior which takes two targets (the ball and the goal, named 
"yo" and "bob" for some reason).  In fact, you could have a sub-behavior which takes THREE
targets, the ball, the goal, and the goal AGAIN, named "no", "way", and "jose".  That's
just fine.

Now, just because an HFA uses a targeted behavior doesn't mean it needs to define targets
itself.  For example, we might provide the targets directly.  For example, imagine that
you had a target called TheGoal stored as a global variable.  You could pass it in directly
as "X" when calling goto like this:

attack = makeHFA("attack", makeTransition(
    {
    [start] = forward,
    [forward] = function(hfa)    
        if (farFromBall()) then return { [0] = goto, ["X"] = TheGoal }
        else return forward
        end,
    [goto] = function(hfa)
        if (closeToBall()) then return forward
        else return { [0] = goto, ["X"] = TheGoal }
        end,
    [kick] = { [0] = goto, ["X"] = TheGoal }, 
    ...
    }))

We call this binding the target to a GROUND VALUE.  Now perhaps attack only needs to be provided
with a "ball" target.  Or if you bound both "goal" and "ball" to ground values when passing them
into sub-behaviors, you wouldn't need either of them in your HFA at all.

Furthermore, your transition functions can pass in different things at different times: you might
have a transition function which provides "goal" as "X", and another one which for some reason
provides "ball" as "X", etc.

So in general, the mapping now looks like this:

{ [0] = the behavior,
        ["behavior's name for first target"] = "hfa's name for target" --OR-- theTargetItself,
        ["behavior's name for second target"] = "hfa's name for second target" --OR-- theTargetItself, 
        ... }

Note that the HFA does not store its targets at all -- it's just up to the transition
functions to map them.  If the targets aren't passed into the HFA by its parent, and you
need them in some sub-behavior, you'll get a mapping error when a transition function tries
to map a target that doesn't exist.

Thus if your HFA is using sub-behaviors which require targets, the transition functions of the
HFA must either bind them to ground values, or they must bind them to targets of their own.
Ultimately, the targets have to be bound to *some* ground value somewhere in the chain of
HFA parents, clear up to pulse().  This is why pulse takes a mapping of target names to targets:

pulse(attack, { ["goal"] = TheGoal, ["ball"] = TheBall })

Of course, if attack doesn't need targets, you can just pass in nil or an empty table.

Additionally, your HFA's transition function may find use of a target helpful.  Perhaps you need
to know the distance to the ball in order to return the right behavior.  In an HFA, the target table
(of the form { ["targetName"] = ground target , ... } ) is stored here:

myHFA.targets

Similarly, the targets, as mapped for the current state, are stored here:

myHFA.behaviorTargets


(You're less likely to use this one)

Thus you could do something like this:

    [forward] = function(hfa)
        ball = hfa.targets.ball
        if (farFrom(ball)) then return{ [0] = goto, ["X"] = TheGoal }
        else return forward
        end,
    
So now we have TWO reasons why an HFA may require targets passed to it:

    1. Because some sub-behavior requires a target and the HFA isn't binding the
       target to a ground value, so the HFA itself needs the target passed in (perhaps
       as a different name)
       
    2. Because the HFA uses the target in a transition function to compute
       the proper transition.
       

KEEP IN MIND that there are two kinds of tables involved here, which look similar but
are different:

    - TARGET TABLES { ["target name"] = ground target, ... }
      Are passed into the START, STOP, and GO methods of behaviors and are also provided
      to the pulse() method.
      
    - MAPPINGS { [0] = behavior, ["target name in behavior"] = "target name in parent HFA"
                                 --OR-- ["target name in behavior"] = ground target, ... }
      Are returned by transition functions to indicate the behavior to transition to and also
      the target mapping to use when passing targets into that behavior.
      


-- ADDITIONAL STUFF TO KEEP IN MIND --

You should be aware of the fact that this library effectively creates directed acyclic
graphs (DAGs) of HFA.  That is, the exact same behavior may be used by different
parent behaviors.  This could have an effect if the behavior maintains some internal
state which is not reset when STOP or START are called.  For example, imagine you had
a behavior called COUNT:

theCount = 0
count = makeBehavior("count", nil, nil, 
    function(behavior, targets) theCount = theCount + 1; print(theCount) end
    )

Now imagine you had two HFAs which both include this behavior among their states.
You're in HFA #1 and it calls count four times, resulting in the numbers 1...4 being
printed to the screen.  Then you're in HFA #2 and it calls count five times.  This
results in the numbers 5...9 being printed to the screen.  This happens because the
same behavior is being used, not multiple copies of it.  If you want multiple copies
of behaviors, you'll need to make them.  Keep this in mind.



... it's always a good sign when the comments are much longer than the code, right?
]]--



-- makebehavior(name, start, stop, go)
-- Creates a behavior with the given name and start/stop/go functions.  Any
-- of these functions can be nil.
makeBehavior = function(name, start, stop, go)
    return { ["name"] = name, ["start"] = start, ["stop"] = stop, 
    		 ["go"] = go, ["parent"] = nil, ["pulsed"] = false }
end


-- start
-- This is the start state.  It's an empty behavior,
start = makeBehavior("start", nil, nil, nil)


-- translateTargets(targets, mapping)
-- Translates targets in the form of { ["targetName"] = target, ... }
-- Into targets of the form of { ["mappedName"] = target, ... }
-- Using the mapping { [0] = behavior, ["mappedName"] = "targetName", ... }
-- A private internal function used by stopHFA(...) and goHFA(...)
translateTargets = function(targets, mapping)
    mapped = { }
    if (mapping[0] == nil) then
        print("ERROR translateTargets(targets, mapping): mapping does not contain 0 as a key") 
    end
    for mapname, original in pairs(mapping) do
        if (not (mapname == 0)) then
            if (type(original) == "string") then
                if (targets == nil) then
                    print ("ERROR translateTargets(targets, mapping): targets is nil while mapping")
                elseif (targets[original] == nil) then
                    print ("WARNING translateTargets(targets, mapping): targets[original] is nil while mapping")
                else
                    mapped[mapname] = targets[original]
                end
            else
                -- we assume it's a ground target
                mapped[mapname] = original
            end
        end
    end
    return mapped
end



-- postEvent(hfa, event)
-- posts an event to the HFA.  This may or may not cause the transition function
-- in the HFA to transition to a new state next iteration.  Events are cleared
-- after each new iteration.
postEvent = function(hfa, event)
	table.insert(hfa.events, event)
end

-- clearEvents(hfa)
-- clears all events in the HFA.
clearEvents = function(hfa)
	hfa.events = { }
end

-- setReturnEvent(hfa, event)
-- sets the event that the HFA will return at the conclusion of its go(...) method being
-- called.  This could trigger its *parent* HFA, if any, to perform some kind of transition.
-- If event is nil, then the HFA returns nil instead.
setReturnEvent = function(hfa, event)
	hfa.returnEvent = event
end



-- startHFA(hfa, targets)
-- Private internal function which is the START function for an HFA.
startHFA = function(hfa, targets)
    hfa.done = false;
    hfa.failed = false;
    hfa.counter = 0;
    -- maybe this is too costly and we should restrict it to the resetTimer function?
    hfa.timer = os.time()
    hfa.goReturned = nil;
    hfa.current = start;
end

-- stopHFA(hfa)
-- Private internal function which is the STOP function for an HFA.
stopHFA = function(hfa, targets)
	hfa.targets = nil
    if (hfa.current == nil) then
    	print("WARNING (stopHFA) current is nil")
	else
		if (not (hfa.current.stop == nil)) then
   		     hfa.current.stop(hfa.current, hfa.behaviorTargets)
    	end
    	hfa.current.parent = nil
    end
	clearEvents(hfa) 
end

-- goHFA(hfa)
-- Private internal function which is the GO function for an HFA.
goHFA = function(hfa, targets)
    hfa.targets = targets

    if (not (hfa.goReturned == nil)) then
    	postEvent(hfa, hfa.goReturned)
    	hfa.goReturned = nil
    end

    -- DETERMINE TRANSITION
    local newBehavior = nil
    
    if (not (hfa.transition == nil)) then
        newBehavior = hfa.transition(hfa)
    end
    if (newBehavior == hfa.current) then
    	newBehavior = nil
	end
	
	clearEvents(hfa)    		
    
    -- EXTRACT TARGETS
    -- this is the previous subset, we use it in calling stop() if necessary
    local oldBehaviorTargets = hfa.behaviorTargets
    -- figure out the new subset.  We test for target lists based on whether
    -- the table provided has 0 as a key
    if (not (newBehavior == nil) and not (newBehavior[0] == nil)) then
        -- translate using newBehavior as a mapping, ignoring [0], which is the "real" newbehavior
        hfa.behaviorTargets = translateTargets(targets, newBehavior)
        newBehavior = newBehavior[0]
    else
        hfa.behaviorTargets = nil
    end
        
    -- PERFORM TRANSITION
    if (not (newBehavior == nil)) then
        if (not (newBehavior == hfa.current)) then
        	if (hfa.current == nil)  then
    			print("WARNING (goHFA) current is nil")
    		else
    			if (not (hfa.current.stop == nil)) then
               	  hfa.current.stop(hfa.current, oldBehaviorTargets)
    			end 
    			hfa.current.parent = nil
    		end
            newBehavior.parent = hfa
			if  (not (newBehavior.start == nil)) then
                newBehavior.start(newBehavior, hfa.behaviorTargets)
            end
            hfa.current = newBehavior
            hfa.done = false
            hfa.failed = false
        end
    else
        print("WARNING (goHFA): nil new behavior")
    end
            
    -- EXECUTE
    if ((not (hfa.current == nil)) and
	    (not (hfa.current.go == nil))) then
        hfa.goReturned = hfa.current.go(hfa.current, hfa.behaviorTargets)
    end
    
    -- RETURN THE RETURNEVENT IF ANY
    if (not (hfa.returnEvent == nil)) then
    	return hfa.returnEvent(hfa, targets)
    else
    	return nil
    end
end


-- event(transitions)
-- Produces a transition function based on the provided transitions.
-- transitions is a table of the form { [event]=state/transitionfunc, ... },
-- where the values can be states or themselves transition functions  
-- The transition function walks through hfa.events from first to last,
-- looking for an event which matches a key in the transitions table.
-- if it finds one, it returns the corresponding state, or if it is
-- a transition function, calls the function (passing in hfa) and returns
-- the result.  If it does not find any matching event, it returns nil 
-- (indicating no transition).
event = function(transitions)
    return function(hfa)
		for event in ipairs(hfa.events) do
			local transition = transitions[event]
			if (not (transition == nil)) then
				if (type(transition) == "function") then
					return transition(hfa)
				else
					return transition
				end
			end
		end
		return nil
    end
end

-- makeTransition(transitions)
-- transitions is a dictionary whose keys are
-- behaviors (states) and whose values are either OTHER behaviors
-- or transition functions of the form foo(hfa) which, when called,
-- return what the next state should be.  If the value is a behavior (state),
-- then the HFA will ALWAYS transition to this second state when it is currently in
-- the state associated with key.  If the value is a transition function, then the
-- HFA will call the transition function and transition to the state indicated.
makeTransition = function(transitions)
    return function(hfa)
        local transition = nil
        if (hfa.current == nil) then
        	print("WARNING (makeTransition): current is nil")
        else
            transition = transitions[hfa.current]
        end
        if (type(transition) == "function") then
            return transition(hfa)
        else
            return transition
        end
    end
end

-- makeHFA(name, transition)
-- Creates an HFA with the given name and transition function.
-- Though the transition function can be nil, it's almost certainlly not appropriate to do so.
makeHFA = function(name, transition)
    return { ["name"] = name, ["start"] = startHFA, ["stop"] = stopHFA, ["go"] = goHFA, 
             ["transition"] = transition, ["pulsed"] = false, ["returnEvent"] = nil,
             ["parent"] = nil, ["goReturned"] = nil, ["counter"] = 0, ["timer"] = 0, 
             ["done"] = false, ["failed"] = false, ["current"] = start, ["events"] = { },
             ["propagateFlags"] = false, ["targets"] = nil, ["behaviorTargets"] = nil }
end

-- makeWrapper(name, wrappedBehavior)
-- Creates a wrapper HFA for the given behavior, which works exactly like the
-- behavior does.  This allows you to have the same behavior appear as multiple
-- states in a parent HFA.
makeWrapper = function(name, wrappedBehavior)
  return makeHFA(name, 
                 function(hfa)
                   return wrappedBehavior 
                 end)
end



-- Pulses the behavior.  This is the top-level stepping procedure for your HFA.
pulse = function(behavior, targets)
    if (not behavior.pulsed) then
        behavior.pulsed = true
        if (not (behavior.start == nil)) then
            behavior.start(behavior, targets)
        end
    end
    if (not (behavior.go == nil)) then
        behavior.go(behavior, targets)
    end
end
        

-- Resets a behavior so that next time it is pulsed, it will call start() again.
reset = function(behavior, targets)
    if (behavior.pulsed) then
        if (not (behavior.stop == nil)) then
            behavior.stop(behavior, targets)
        end
        behavior.pulsed = false
    end
end


-- UTILITY BEHAVIORS --

-- bumpCounter: increments the HFA's counter by 1
bumpCounter = makeBehavior("bumpCounter", 
    function(behavior, targets) 
        if (not (behavior == nil) and not (behavior.parent == nil)) then 
            behavior.parent.counter = behavior.parent.counter + 1
        end
    end, nil, nil)

-- zeroCounter: sets the HFA's counter to 0
resetCounter = makeBehavior("resetCounter",
    function(behavior, targets) 
        if (not (behavior == nil) and not (behavior.parent == nil)) then 
            behavior.parent.counter = 0
        end
    end, nil, nil)
    
-- currentCounter(hfa) returns the HFA's current counter value.  This is a 
--                     helpful utility function for making transition functions.
currentCounter = function(hfa) return hfa.counter end

-- resetTimer: resets the HFA's current timer value.  This is a helpful utility
--                   function for making transition functions.
resetTimer = makeBehavior("resetTimer",
    function(behavior, targets)
        if (not (behavior == nil) and not (behavior.parent == nil)) then
            behavior.parent.timer = os.time()
        end
    end, nil, nil)

-- currentTimer(hfa) returns the difference between the current time and the
--                   HFA's current timer value.  This is a helpful utility
--                   function for making transition functions.
currentTimer = function(hfa) return os.time() - hfa.timer end

-- setFlag(hfa, flag).  Internal private function used by various flag behaviors
--                      to set a flag in an HFA, potentially recursively if
--                      propagation is turned on.
setFlag = function(hfa, flag)
    if (not (hfa.parent == nil)) then
        hfa.parent[flag] = true
        if (hfa.parent.propagateFlags == true) then
            setFlag(hfa.parent, flag)
        end
    end
end

-- done: sets the "done" flag in the HFA's parent, and transitions to "start"
done = makeBehavior("done", nil, nil,
    function(behavior, targets) 
	if (not (behavior == nil) and not (behavior.parent == nil)) then
		behavior.parent.current = start
        setFlag(behavior.parent, "done")
        behavior.parent = nil
    end
 end)

-- sayDone: sets the "done" flag in the HFA's parent
sayDone = makeBehavior("sayDone", 
    function(behavior, targets)
        if (not (behavior == nil) and not (behavior.parent == nil)) then
            setFlag(behavior.parent, "done")
        end
    end, nil, nil)
    
-- isDone(hfa)	returns the current done flag in the HFA
isDone = function(hfa) return hfa.done end

-- failed: sets the "failed" flag in the HFA's parent, and transitions to "start"
failed = makeBehavior("failed", nil, nil,
    function(behavior, targets) 
        if (not (behavior == nil) and not (behavior.parent == nil)) then
            behavior.current = start
            setFlag(behavior.parent, "failed")
            behavior.parent = nil
        end
    end)

-- sayFailed: sets the "failed" flag in the HFA's parent
sayFailed = makeBehavior("sayFailed", 
    function(behavior, targets)
        if (not (behavior == nil) and not (behavior.parent == nil)) then
            setFlag(behavior.parent, "failed")
        end
    end, nil, nil)

-- isFailed(hfa)	returns the current failed flag in the HFA
isFailed = function(hfa) return hfa.failed end
    


-- END HFA.LUA



--[[
  Apache License
  Version 2.0, January 2004
  http://www.apache.org/licenses/

  TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION

  1. Definitions.

  "License" shall mean the terms and conditions for use, reproduction,
  and distribution as defined by Sections 1 through 9 of this document.

  "Licensor" shall mean the copyright owner or entity authorized by
  the copyright owner that is granting the License.

  "Legal Entity" shall mean the union of the acting entity and all
  other entities that control, are controlled by, or are under common
  control with that entity. For the purposes of this definition,
  "control" means (i) the power, direct or indirect, to cause the
  direction or management of such entity, whether by contract or
  otherwise, or (ii) ownership of fifty percent (50%) or more of the
  outstanding shares, or (iii) beneficial ownership of such entity.

  "You" (or "Your") shall mean an individual or Legal Entity
  exercising permissions granted by this License.

  "Source" form shall mean the preferred form for making modifications,
  including but not limited to software source code, documentation
  source, and configuration files.

  "Object" form shall mean any form resulting from mechanical
  transformation or translation of a Source form, including but
  not limited to compiled object code, generated documentation,
  and conversions to other media types.

  "Work" shall mean the work of authorship, whether in Source or
  Object form, made available under the License, as indicated by a
  copyright notice that is included in or attached to the work
  (an example is provided in the Appendix below).

  "Derivative Works" shall mean any work, whether in Source or Object
  form, that is based on (or derived from) the Work and for which the
  editorial revisions, annotations, elaborations, or other modifications
  represent, as a whole, an original work of authorship. For the purposes
  of this License, Derivative Works shall not include works that remain
  separable from, or merely link (or bind by name) to the interfaces of,
  the Work and Derivative Works thereof.

  "Contribution" shall mean any work of authorship, including
  the original version of the Work and any modifications or additions
  to that Work or Derivative Works thereof, that is intentionally
  submitted to Licensor for inclusion in the Work by the copyright owner
  or by an individual or Legal Entity authorized to submit on behalf of
  the copyright owner. For the purposes of this definition, "submitted"
  means any form of electronic, verbal, or written communication sent
  to the Licensor or its representatives, including but not limited to
  communication on electronic mailing lists, source code control systems,
  and issue tracking systems that are managed by, or on behalf of, the
  Licensor for the purpose of discussing and improving the Work, but
  excluding communication that is conspicuously marked or otherwise
  designated in writing by the copyright owner as "Not a Contribution."

  "Contributor" shall mean Licensor and any individual or Legal Entity
  on behalf of whom a Contribution has been received by Licensor and
  subsequently incorporated within the Work.

  2. Grant of Copyright License. Subject to the terms and conditions of
  this License, each Contributor hereby grants to You a perpetual,
  worldwide, non-exclusive, no-charge, royalty-free, irrevocable
  copyright license to reproduce, prepare Derivative Works of,
  publicly display, publicly perform, sublicense, and distribute the
  Work and such Derivative Works in Source or Object form.

  3. Grant of Patent License. Subject to the terms and conditions of
  this License, each Contributor hereby grants to You a perpetual,
  worldwide, non-exclusive, no-charge, royalty-free, irrevocable
  (except as stated in this section) patent license to make, have made,
  use, offer to sell, sell, import, and otherwise transfer the Work,
  where such license applies only to those patent claims licensable
  by such Contributor that are necessarily infringed by their
  Contribution(s) alone or by combination of their Contribution(s)
  with the Work to which such Contribution(s) was submitted. If You
  institute patent litigation against any entity (including a
  cross-claim or counterclaim in a lawsuit) alleging that the Work
  or a Contribution incorporated within the Work constitutes direct
  or contributory patent infringement, then any patent licenses
  granted to You under this License for that Work shall terminate
  as of the date such litigation is filed.

  4. Redistribution. You may reproduce and distribute copies of the
  Work or Derivative Works thereof in any medium, with or without
  modifications, and in Source or Object form, provided that You
  meet the following conditions:

  (a) You must give any other recipients of the Work or
  Derivative Works a copy of this License; and

  (b) You must cause any modified files to carry prominent notices
  stating that You changed the files; and

  (c) You must retain, in the Source form of any Derivative Works
  that You distribute, all copyright, patent, trademark, and
  attribution notices from the Source form of the Work,
  excluding those notices that do not pertain to any part of
  the Derivative Works; and

  (d) If the Work includes a "NOTICE" text file as part of its
  distribution, then any Derivative Works that You distribute must
  include a readable copy of the attribution notices contained
  within such NOTICE file, excluding those notices that do not
  pertain to any part of the Derivative Works, in at least one
  of the following places: within a NOTICE text file distributed
  as part of the Derivative Works; within the Source form or
  documentation, if provided along with the Derivative Works; or,
  within a display generated by the Derivative Works, if and
  wherever such third-party notices normally appear. The contents
  of the NOTICE file are for informational purposes only and
  do not modify the License. You may add Your own attribution
  notices within Derivative Works that You distribute, alongside
  or as an addendum to the NOTICE text from the Work, provided
  that such additional attribution notices cannot be construed
  as modifying the License.

  You may add Your own copyright statement to Your modifications and
  may provide additional or different license terms and conditions
  for use, reproduction, or distribution of Your modifications, or
  for any such Derivative Works as a whole, provided Your use,
  reproduction, and distribution of the Work otherwise complies with
  the conditions stated in this License.

  5. Submission of Contributions. Unless You explicitly state otherwise,
  any Contribution intentionally submitted for inclusion in the Work
  by You to the Licensor shall be under the terms and conditions of
  this License, without any additional terms or conditions.
  Notwithstanding the above, nothing herein shall supersede or modify
  the terms of any separate license agreement you may have executed
  with Licensor regarding such Contributions.

  6. Trademarks. This License does not grant permission to use the trade
  names, trademarks, service marks, or product names of the Licensor,
  except as required for reasonable and customary use in describing the
  origin of the Work and reproducing the content of the NOTICE file.

  7. Disclaimer of Warranty. Unless required by applicable law or
  agreed to in writing, Licensor provides the Work (and each
  Contributor provides its Contributions) on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
  implied, including, without limitation, any warranties or conditions
  of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
  PARTICULAR PURPOSE. You are solely responsible for determining the
  appropriateness of using or redistributing the Work and assume any
  risks associated with Your exercise of permissions under this License.

  8. Limitation of Liability. In no event and under no legal theory,
  whether in tort (including negligence), contract, or otherwise,
  unless required by applicable law (such as deliberate and grossly
  negligent acts) or agreed to in writing, shall any Contributor be
  liable to You for damages, including any direct, indirect, special,
  incidental, or consequential damages of any character arising as a
  result of this License or out of the use or inability to use the
  Work (including but not limited to damages for loss of goodwill,
  work stoppage, computer failure or malfunction, or any and all
  other commercial damages or losses), even if such Contributor
  has been advised of the possibility of such damages.

  9. Accepting Warranty or Additional Liability. While redistributing
  the Work or Derivative Works thereof, You may choose to offer,
  and charge a fee for, acceptance of support, warranty, indemnity,
  or other liability obligations and/or rights consistent with this
  License. However, in accepting such obligations, You may act only
  on Your own behalf and on Your sole responsibility, not on behalf
  of any other Contributor, and only if You agree to indemnify,
  defend, and hold each Contributor harmless for any liability
  incurred by, or claims asserted against, such Contributor by reason
  of your accepting any such warranty or additional liability.

  END OF TERMS AND CONDITIONS

  APPENDIX: How to apply the Apache License to your work.

  To apply the Apache License to your work, attach the following
  boilerplate notice, with the fields enclosed by brackets "[]"
  replaced with your own identifying information. (Don't include
  the brackets!)  The text should be enclosed in the appropriate
  comment syntax for the file format. We also recommend that a
  file or class name and description of purpose be included on the
  same "printed page" as the copyright notice for easier
  identification within third-party archives.

  Copyright [yyyy] [name of copyright owner]

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
]]--